package ru.dbotthepony.mc.otm.menu

import net.minecraft.network.chat.Component
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.client.render.sprites.AbstractMatterySprite
import ru.dbotthepony.mc.otm.container.IFilteredContainerSlot
import ru.dbotthepony.mc.otm.container.util.containerSlotOrNull
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.util.ItemStackKey
import ru.dbotthepony.mc.otm.core.util.asKey
import ru.dbotthepony.mc.otm.core.util.asKeyOrNull

class QuickMoveInput(private val menu: MatteryMenu, val from: Collection<Slot>, val to: Collection<Slot>, val mode: Mode) {
	enum class Mode(val iconFromStorage: AbstractMatterySprite, val iconToStorage: AbstractMatterySprite) {
		RESTOCK(
			Widgets18.RESTOCK_FROM_STORAGE,
			Widgets18.RESTOCK_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player) {
				if (from.isEmpty() || to.isEmpty()) return
				val (_, itemsFrom) = computeSlotLists(from, true)
				val (_, itemsTo) = computeSlotLists(to, false)

				val intersect = if (itemsFrom.size < itemsTo.size) itemsFrom.keys.filter { it in itemsTo.keys } else itemsTo.keys.filter { it in itemsFrom.keys }

				for (key in intersect) {
					val slotsTo = itemsTo[key]!!
					itemsFrom[key]!!.forEach { moveItemStackTo(player, it, slotsTo) }
				}
			}
		},

		RESTOCK_WITH_MOVE(
			Widgets18.RESTOCK_WITH_MOVE_FROM_STORAGE,
			Widgets18.RESTOCK_WITH_MOVE_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player) {
				if (from.isEmpty() || to.isEmpty()) return
				val (_, itemsFrom) = computeSlotLists(from, true)
				val (emptyTo, itemsTo) = computeSlotLists(to, false)

				val intersect = if (itemsFrom.size < itemsTo.size) itemsFrom.keys.filter { it in itemsTo.keys } else itemsTo.keys.filter { it in itemsFrom.keys }

				for (key in intersect) {
					val slotsTo = prioritySortSlots(itemsTo[key]!!, key.asItemStack())
					val slotsFrom = itemsFrom[key]!!
					slotsFrom.removeIf { moveItemStackTo(player, it, slotsTo, sort = false); it.item.isEmpty }
					var moveAny = false
					slotsFrom.forEach { moveAny = moveItemStackTo(player, it, emptyTo, sort = false) || moveAny }
					if (moveAny) emptyTo.removeIf { it.item.isNotEmpty }
				}
			}
		},

		MOVE(
			Widgets18.MOVE_EVERYTHING_FROM_STORAGE,
			Widgets18.MOVE_EVERYTHING_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player) {
				if (from.isEmpty() || to.isEmpty()) return
				val toSorted = prioritySortSlots(to)

				from.forEach {
					val slot = it.containerSlotOrNull()

					if (slot !is IFilteredContainerSlot || !slot.hasFilter)
						moveItemStackTo(player, it, toSorted, sort = false)
				}
			}
		};

		abstract fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player)

		val textFromStorage: Component get() {
			return TranslatableComponent("otm.gui.quickmove_from.${name.lowercase()}")
		}

		val textToStorage: Component get() {
			return TranslatableComponent("otm.gui.quickmove_to.${name.lowercase()}")
		}
	}

	val input = menu.oneWayInput(handler = ::handle)

	fun clientInput() {
		input.accept(null)
	}

	private fun handle() {
		mode.move(from, to, menu.player)
	}

	companion object {
		fun create(menu: MatteryMenu, from: Collection<Slot>, to: Collection<Slot>): Map<Mode, QuickMoveInput> {
			return Mode.entries.associateWith { QuickMoveInput(menu, from, to, it) }
		}

		private fun computeSlotLists(slots: Collection<Slot>, skipFilteredSlots: Boolean): Pair<MutableList<Slot>, MutableMap<ItemStackKey, MutableList<Slot>>> {
			val emptySlots = ArrayList<Slot>()
			val filledSlots = HashMap<ItemStackKey, MutableList<Slot>>()

			for (slot in slots) {
				val underlyingSlot = slot.containerSlotOrNull()

				if (underlyingSlot is IFilteredContainerSlot && (underlyingSlot.filter == Items.AIR || underlyingSlot.filter != null && skipFilteredSlots))
					continue

				val key = slot.item.asKeyOrNull() ?: (underlyingSlot as? IFilteredContainerSlot)?.filter?.asKey()

				if (key == null) {
					emptySlots.add(slot)
				} else {
					filledSlots.computeIfAbsent(key) { ArrayList() }.add(slot)
				}
			}

			return emptySlots to filledSlots
		}

		fun moveItemStackTo(
			player: Player,
			source: Slot,
			slots: Collection<Slot>,
			sort: Boolean = true
		): Boolean {
			if (!source.mayPickup(player) || source.item.isEmpty || slots.isEmpty())
				return false

			val remainder = moveItemStackToSlots(source.item, slots, sort = sort)

			if (remainder.count == source.item.count)
				return false

			val copy = source.item.copy()

			if (remainder.isEmpty) {
				source.setByPlayer(ItemStack.EMPTY)
				source.onTake(player, copy)
			} else {
				copy.count = source.item.count - remainder.count
				source.item.count = remainder.count
				source.onTake(player, copy)
			}

			return true
		}

		fun prioritySortSlots(slots: Collection<Slot>, filterItem: ItemStack? = null): MutableList<Slot> {
			val sortedSlots = ArrayList(slots)

			sortedSlots.removeIf {
				val slot = it.containerSlotOrNull()
				it.isOverCapacity || filterItem != null && !it.mayPlace(filterItem) || slot is IFilteredContainerSlot && slot.isForbiddenForAutomation
			}

			sortedSlots.sortWith { a, b ->
				val hasItemA = a.item.isNotEmpty
				val hasItemB = b.item.isNotEmpty

				if (hasItemA && hasItemB)
					return@sortWith 0
				else if (hasItemA)
					return@sortWith -1
				else if (hasItemB)
					return@sortWith 1

				val slotA = a.containerSlotOrNull()
				val slotB = b.containerSlotOrNull()

				val hasFilterA = slotA is IFilteredContainerSlot && slotA.hasFilter
				val hasFilterB = slotB is IFilteredContainerSlot && slotB.hasFilter

				if (hasFilterA && hasFilterB || !hasFilterA && !hasFilterB)
					return@sortWith 0
				else if (hasFilterA)
					return@sortWith -1
				else
					return@sortWith 1
			}

			return sortedSlots
		}

		fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false, sort: Boolean = true): ItemStack {
			if (item.isEmpty)
				return ItemStack.EMPTY
			else if (slots.isEmpty())
				return item.copy()

			val sortedSlots = if (sort) prioritySortSlots(slots, item) else slots
			val copy = item.copy()

			for (slot in sortedSlots) {
				val limit = slot.getMaxStackSize(copy)

				if (!slot.hasItem()) {
					val newCount = copy.count.coerceAtMost(limit)

					if (!simulate) {
						slot.setByPlayer(copy.copy().also { it.count = newCount })
						// slot.setChanged()
					}

					copy.shrink(newCount)

					if (copy.isEmpty)
						return ItemStack.EMPTY
				} else if (limit > slot.item.count && ItemStack.isSameItemSameComponents(slot.item, copy)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty)
						return ItemStack.EMPTY
				}
			}

			return copy
		}
	}
}
